"""
Logic surrounding templates manipulation
"""
from dataclasses import dataclass, asdict
import enum

@dataclass
class FunctionTemplateParams:
    """
    Standardises which paramters need to be passed
    to render a 'function'template
    """
    function_name: str
    function_body: str 

class MetaDictionary(enum.Enum):
    """
    this are the keywork that we map every
    wrapper to for translationj
    """
    RECEIVED_MESSAGE = 'input message string for the bot function'
    SEND_MESSAGE = 'method to send a message as reply'
    CHAT = 'chat object'
    USER_SENDER = 'user object who sent the message'



import enum
DEFAULT_OUTPUT_FILENAME = "gen__xbot.py"
GENERATED_COMMENT = '"""__GENERATED__ cross platform generated function"""'

class TEMPLATES(enum.Enum):
    """
    These are the supported templates, i.e. the classes
    of function that xbot can generate
    """

    # REPLY is basic get arguments, 
    # process them and send reply template
    REPLY = 'reply.py'

class LIBRARIES(enum.Enum):
    """
    These are the supported messaing platform APIs
    wrapper to translate between
    """

    # https://github.com/Rapptz/discord.py
    DISCORD_PY = 'discord'

    # https://github.com/python-telegram-bot/python-telegram-bot
    PYTHON_TELEGRAM_BOT = 'telegram'



"""core e2e translation logic"""
import ast
import logging
from typing import Iterator
from  dataclasses import asdict

import xbot.constants
import xbot.utils
import xbot.templates

def parse_source_code(
        filename: str, 
        from_library = xbot.constants.LIBRARIES.PYTHON_TELEGRAM_BOT,
        to_library = xbot.constants.LIBRARIES.DISCORD_PY
        ) -> Iterator[xbot.templates.FunctionTemplateParams]:
    """
    Translates source code written in `from_library` into a 
    temporary meta-language, using MetaDictionary and Templates.
    """
    logging.info("parse_source_code..")
    sourcecode = ""
    with open(filename, "r") as f:
        sourcecode = f.read()

    Parser = xbot.utils.get_parser(from_library)
    dictionary = xbot.utils.make_dictionary(from_library, to_library)

    parser = Parser(sourcecode, dictionary)
    translated_functions = parser.parse()
    return translated_functions

def generate_destination_code(
        bot_functions: Iterator[xbot.templates.FunctionTemplateParams], 
        output_file: str,
        output_library: xbot.constants.LIBRARIES
        ):
    """
    Translates the temporary meta-language in
    output_library executable code.
    """
    logging.info("generate_destination_code..")

    template = xbot.utils.template_filename(
            output_library,
            xbot.constants.TEMPLATES.REPLY
            )
    translated_functions = map(
            lambda params: 
                xbot.utils.render_jinja_template(template, asdict(params)),
            bot_functions
            )
    translated_code = "\n\n".join(translated_functions)

    logging.info("\n\n"+translated_code)
    with open(output_file, "w") as f:
        f.write(translated_code)

def translate(
        sourcecode_filename, 
        input_library=xbot.constants.LIBRARIES.PYTHON_TELEGRAM_BOT,
        output_file=None,
        output_library=xbot.constants.LIBRARIES.DISCORD_PY
        ):
    """
    entrypoint for the translation 
    """
    logging.info("translate {}..".format(sourcecode_filename))
    params = parse_source_code(sourcecode_filename)

    if output_file is None or output_file == "":
        output_file = xbot.constants.DEFAULT_OUTPUT_FILENAME
    generate_destination_code(
            params, 
            output_file,
            output_library
            )



"""xbot entrypoint"""
import argparse
import logging
import xbot.constants
import xbot.translate

def parse_commandline_args():
    parser = argparse.ArgumentParser(description='xbot translates your bot code from one messaging platform to another, generates a file gen__xbot.py or --output_file')
    parser.add_argument('sourcecode_filename', metavar='sourcecode_filename', type=str, nargs=1,
                        help='name of the file to translate')
    parser.add_argument('--source_library', type=str,
                        default="python-telegram-bot",
                        help='library in which the source code is written in')
    parser.add_argument('--destination_library', type=str,
                        default="discord.py",
                        help='library to translate the code into')
    parser.add_argument('--output_file', type=str,
                        default="",
                        help='name of the output filee')
    parser.add_argument('--verbose', 
                        action='store_true',
                        help='verbose logging')
    parser.add_argument('-v', 
                        action='store_true',
                        help='verbose logging')
    args = parser.parse_args()
    return args

if __name__ == "__main__":
    args = parse_commandline_args()
    if args.verbose or args.v:
        logging.basicConfig(level=logging.INFO)
    xbot.translate.translate(args.sourcecode_filename[0], output_file=args.output_file)



import xbot.templates
import xbot.constants
import logging
from typing import Iterator
import ast
import astunparse

class Parser:
    """
    The python-telegram-bot API source code must follow
    this guidelines

    - all module level handlers
    - all the telegram code used must be in the dictionary provided

    You can easily construct any language to language dictionary
    using MetaDictionary
    """
    def __init__(
            self, 
            sourcecode: str, 
            dictionary: dict,
            output_indentation_level: int = 1,
            output_indentation_string: str = "    "
            ):
        self.sourcecode = sourcecode
        self.dictionary = dictionary
        self.indentation_level = output_indentation_level
        self.indentation_string = output_indentation_string
        self.parse()

    def _get_bot_functions(self, ast_module: ast.Module):
        """returns functions from the first level of the AST"""
        module_level_functions = filter(
                lambda node: type(node) == ast.FunctionDef, ast_module.body)

        decorator_list = lambda ast_function: \
            ast_function.decorator_list \
            if 'decorator_list' in ast_function._fields else []

        has_xbot_decorator = lambda ast_function: \
                any(map(lambda ast_decorator: ast_decorator.value.id == 'xbot', 
                    decorator_list(ast_function)))

        xbot_decorated_functions = filter(has_xbot_decorator, module_level_functions)
        return xbot_decorated_functions

    def _translate_line(self, original_line: str) -> str:
        """replace tokens from self.dictionary when found"""
        translated_line = original_line
        for original_token, translated_token in self.dictionary.items():
            translated_line = translated_line\
                    .replace(original_token, translated_token)
        # indent code at every \n
        translated_line = translated_line[:-1].replace( 
                "\n", "\n"+self.indentation_level * self.indentation_string)
        return translated_line

    def _get_function_body(self, ast_function: ast.FunctionDef)  -> str:
        """proceding line by line, unparse the function body and translates it"""
        original_code = map(lambda node: astunparse.unparse(node), ast_function.body)
        translated_code = map(lambda line: self._translate_line(line), original_code)
        return xbot.constants.GENERATED_COMMENT + "\n".join(translated_code)

    def _get_function_params(self, ast_function: ast.FunctionDef)  -> xbot.templates.FunctionTemplateParams:
        """extracts FunctionTemplateParams for an ast.FunctionDef"""
        return xbot.templates.FunctionTemplateParams(
                function_name = ast_function.name,
                function_body = self._get_function_body(ast_function)
                )

    def parse(self) -> Iterator[xbot.templates.FunctionTemplateParams]:
        ast_root = ast.parse(self.sourcecode)
        functions = self._get_bot_functions(ast_root)
        return map(self._get_function_params, functions)




"""
Dictionary to map telegram code to MetaDictionary
"""
from xbot.templates import MetaDictionary

library_to_meta = {
    'update.message.text': MetaDictionary.RECEIVED_MESSAGE,
    'update.message.reply_text': MetaDictionary.SEND_MESSAGE,
    'update.message.chat': MetaDictionary.CHAT,
    'update.message.from_user': MetaDictionary.USER_SENDER,
}

meta_to_library = { value: key for key, value 
        in library_to_meta.items() }



import functools

def xfunction(func):
    """
    This is a decorator to add in front of functions
    you want to translate. It doesn't do anything by
    itself but it's used by the parser to figure out
    whether to translate this function.
    """
    @functools.wraps(func)
    def wrapper(*args, **kwds):
        return func(*args, **kwds)
    return wrapper




"""
Dictionary to map discord code to MetaDictionary
"""
from xbot.templates import MetaDictionary

library_to_meta = {
    'ctx.message.content': MetaDictionary.RECEIVED_MESSAGE,
    'await ctx.send': MetaDictionary.SEND_MESSAGE,
    'ctx.message.channel': MetaDictionary.CHAT,
    'ctx.message.author': MetaDictionary.USER_SENDER,
}

meta_to_library = { value: key for key, value 
        in library_to_meta.items() }



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="x-bot", 
    version="0.0.5",
    author="Alessandro Solbiati",
    author_email="alessandro.solbiati@gmail.com",
    description="code generator for messaging bot in multiple platforms ",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/SolbiatiAlessandro/xbot",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        'Jinja2',
        'astunparse',
        'pytest'
        ],
    python_requires='>=3.6',
)



import jinja2
import os
import importlib
import logging

import xbot.constants

def render_jinja_template(filename: str, args: dict = {}) -> str:
    template = ""
    with open(filename) as f:
        template = f.read()
    jinja_template = jinja2.Template(template)
    translated_code = jinja_template.render(args)
    return translated_code

def template_filename(
        library: xbot.constants.LIBRARIES, 
        template: xbot.constants.TEMPLATES
        ) -> str:
    return os.path.join("xbot", "templates", library.value, template.value)

def get_parser(library: xbot.constants.LIBRARIES):
    """
    return a Parser class for a given library
    """
    logging.info("get_parser..")
    parser_path = ".".join(["xbot", "parsers", library.value, "parser"])
    parser_module = importlib.import_module(parser_path, "xbot")

    parser = parser_module.Parser
    logging.info(parser)
    return parser

def make_dictionary(
        from_library: xbot.constants.LIBRARIES,
        to_library: xbot.constants.LIBRARIES
        ):
    """
    uses MetaDictionary to dinamically construct a 1:1 
    mapping for the two libraries
    """
    logging.info("make_dictionary..")
    from_library_path = ".".join(
            ["xbot", "parsers", from_library.value, "dictionary"])
    from_library_module = importlib.import_module(from_library_path, "xbot")
    from_library_to_meta = from_library_module.library_to_meta

    to_library_path = ".".join(
            ["xbot", "parsers", to_library.value, "dictionary"])
    to_library_module = importlib.import_module(to_library_path, "xbot")
    meta_to_library = to_library_module.meta_to_library

    from_library_to_library = {
            from_library_key: meta_to_library[meta_key]
            for from_library_key, meta_key in from_library_to_meta.items()
            }

    logging.info(from_library_to_library)
    return from_library_to_library

